/* Copyright (c) 2012 imacat
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * CalcMosaic
 *
 * Created on 2012-08-12
 * 
 * Copyright (c) 2012 imacat
 */

package tw.idv.imacat.calcmosaic;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

/**
 * A program to create mosaic art from images with Calc.
 *
 * @author <a href="mailto:imacat&#64;mail.imacat.idv.tw">imacat</a>
 * @version 0.2.1
 */
public class CalcMosaic {
    
    /** The version number. */
    private final static String VERSION = "0.2.0";
    
    /** The version number. */
    private final static String AUTHOR = "imacat <imacat@mail.imacat.idv.tw>";
    
    /** The version number. */
    private final static String JAR_NAME = "calcmosaic.jar";
    
    /** The localization resources. */
    private ResourceBundle l10n = null;
    
    /** The image files. */
    private List<File> files = new ArrayList<File>();
    
    /**
     * The maximum number of rows to use, default to 0 (estimated
     * automatically).
     */
    private int rows = 0;
    
    /**
     * Creates a new instance of CalcMosaic.
     * 
     * @param args the command line arguments
     */
    public CalcMosaic(String args[]) {
        // Obtains the localization resources
        this.l10n = ResourceBundle.getBundle(
            this.getClass().getPackage().getName() + ".res.L10n");
        this.parseArguments(args);
    }
    
    /**
     * Runs CalcMosaic from the command line.
     * 
     * @param args the command line arguments
     * @throws com.sun.star.comp.helper.BootstrapException if fails to
     *         create the initial component context
     * @throws java.io.IOException when file couldn't be found or
     *         was corrupt
     */
    public static void main(String args[])
            throws com.sun.star.comp.helper.BootstrapException,
                java.io.IOException {
        CalcMosaic mosaic = new CalcMosaic(args);
        
        try {
            mosaic.create();
        } catch (com.sun.star.comp.helper.BootstrapException e) {
            throw e;
        } catch (java.io.IOException e) {
            throw e;
        }
        System.exit(0);
    }
    
    /**
     * Parses the command line arguments.
     * 
     * @param args the command line arguments
     * @throws IllegalArgumentException if the arguments are
     *         illegal
     */
    private void parseArguments(String args[])
            throws IllegalArgumentException {
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-r") || args[i].equals("-rows")) {
                i++;
                if (i >= args.length) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify the cell size.");
                    throw new IllegalArgumentException(JAR_NAME + ": "
                        + "Please specify the cell size.");
                }
                this.rows = Integer.parseInt(args[i]);
                if (rows <= 0) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify a positive number of rows.");
                    throw new IllegalArgumentException(
                        "Please specify a positive number of rows to use.");
                }
            } else if (args[i].startsWith("-r=")) {
                this.rows = Integer.parseInt(args[i].substring(3));
                if (rows <= 0) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify a positive number of rows.");
                    throw new IllegalArgumentException(
                        "Please specify a positive number of rows to use.");
                }
            } else if (args[i].startsWith("-rows=")) {
                this.rows = Integer.parseInt(args[i].substring(6));
                if (rows <= 0) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify a positive number of rows.");
                    throw new IllegalArgumentException(
                        "Please specify a positive number of rows to use.");
                }
            } else if (args[i].equals("-h") || args[i].equals("-help")) {
                System.out.println(
"Usage: java -jar " + JAR_NAME + " [options] file1 [file2...]\n"
+ "Create mosaic art from images with Calc.\n"
+ "\n"
+ "  -r,--rows num    The maximun number of rows to use (default to\n"
+ "                   calculate automatically).\n"
+ "  -h,--help        Display this help.\n"
+ "  -v,--version     Display version number.\n"
+ "  file1, file2...  The image files.\n"
+ "\n"
+ "The snapshots of a video can be grabbed using ffmpeg:\n"
+ "ffmpeg -i video.mp4 -r <framerate> -sameq -y -f image2 img%04d.jpg\n"
+ "Note that a spreadsheet document can only contain at most 256 spreadsheets.");
                System.exit(0);
            } else if (args[i].equals("-v") || args[i].equals("-version")) {
                System.out.println(
JAR_NAME + " v" + VERSION + " by " + AUTHOR);
                System.exit(0);
            } else if (args[i].startsWith("-")) {
                System.err.println(JAR_NAME + ": "
                    + "Unrecognized switch: " + args[i]);
                throw new IllegalArgumentException(
                    "Unrecognized switch: " + args[i]);
            } else {
                files.add(new File(args[i]));
            }
        }
        
        // Checks the limits of the number of images/spreadsheets.
        if (files.size() == 0) {
            System.err.println(JAR_NAME + ": "
                + "Please specify the image files.");
            throw new IllegalArgumentException(
                "Please specify the image files.");
        } else if (files.size() > 256) {
            System.err.println(JAR_NAME + ": "
                + "Too many images (256 max).  A"
                + " spreadsheet document can only contain at most 256"
                + " spreadsheets.");
            throw new IllegalArgumentException(
                "Too many images (256 max).  A spreadsheet document"
                + " can only contain at most 256 spreadsheets.");
        }
    }
    
    /**
     * Creates mosaic effect of images with Calc.
     *
     * @throws com.sun.star.comp.helper.BootstrapException if fails to
     *         create the initial component context
     * @throws java.io.IOException when file couldn't be found or was corrupt
     */
    public void create()
            throws com.sun.star.comp.helper.BootstrapException,
                java.io.IOException {
        Office office = null;
        SpreadsheetDocument document = null;
        
        try {
            office = new Office();
        } catch (com.sun.star.comp.helper.BootstrapException e) {
            throw e;
        }
        document = office.startNewSpreadsheetDocument();
        for (int i = 0; i < files.size(); i++) {
            try {
                document.createMosaic(this.files.get(i), i,
                    this.rows);
            } catch (java.io.IOException e) {
                throw e;
            }
        }
    }
    
    /**
     * Gets a string for the given key from this resource bundle
     * or one of its parents.  If the key is missing, returns
     * the key itself.
     * 
     * @param key the key for the desired string 
     * @return the string for the given key 
     */
    private String _(String key) {
        try {
            return this.l10n.getString(key);
        } catch (java.util.MissingResourceException e) {
            return key;
        }
    }
}
