/* Copyright (c) 2012 imacat
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * SpreadsheetDocument
 *
 * Created on 2012-08-16, rewritten from CalcMosaic
 * 
 * Copyright (c) 2012 imacat
 */

package tw.idv.imacat.calcmosaic;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.ResourceBundle;
import javax.imageio.ImageIO;

import com.sun.star.awt.Size;
import com.sun.star.awt.XView;
import com.sun.star.awt.XWindow;
import com.sun.star.beans.XPropertySet;
import com.sun.star.frame.XController;
import com.sun.star.frame.XFrame;
import com.sun.star.frame.XModel;
import com.sun.star.lang.XComponent;
import com.sun.star.sheet.XSpreadsheet;
import com.sun.star.sheet.XSpreadsheetDocument;
import com.sun.star.sheet.XSpreadsheets;
import com.sun.star.sheet.XSpreadsheetView;
import com.sun.star.table.XCell;
import com.sun.star.table.XColumnRowRange;
import com.sun.star.table.XTableColumns;
import com.sun.star.table.XTableRows;
import com.sun.star.uno.UnoRuntime;

/**
 * A spreadsheet document.
 *
 * @author <a href="mailto:imacat&#64;mail.imacat.idv.tw">imacat</a>
 * @version 0.1.0
 */
public class SpreadsheetDocument {
    
    /** The localization resources. */
    private ResourceBundle l10n = null;
    
    /** The spreadsheet document. */
    private XSpreadsheetDocument document = null;
    
    /** The document controller. */
    private XController xController = null;
    
    /** The spreadsheet view. */
    private XSpreadsheetView xSpreadsheetView = null;
    
    /** The spreadsheet document controller. */
    private Dimension size = null;
    
    /**
     * Creates a new instance of SpreadsheetDocument.
     *
     * @param document the document component
     */
    public SpreadsheetDocument(XComponent document) {
        XModel xModel = null;
        
        // Obtains the localization resources
        this.l10n = ResourceBundle.getBundle(
            this.getClass().getPackage().getName() + ".res.L10n");
        
        this.document = (XSpreadsheetDocument)
            UnoRuntime.queryInterface(
            XSpreadsheetDocument.class, document);
        
        // Obtains the document controller
        xModel = (XModel) UnoRuntime.queryInterface(
            XModel.class, document);
        this.xController = xModel.getCurrentController();
        this.xSpreadsheetView = (XSpreadsheetView)
            UnoRuntime.queryInterface(
            XSpreadsheetView.class, this.xController);
    }
    
    /**
     * Creates a mosaic art spreadsheet from an image file, with
     * mosaic cell size calculated automatically.
     *
     * @param file       the source image file
     * @param sheetIndex the index of the mosaic spreadsheet
     * @throws java.io.IOException when file couldn't be found or
     *         was corrupt
     */
    public void createMosaic(File file, int sheetIndex)
            throws java.io.IOException {
        createMosaic(file, sheetIndex, 0);
    }
    
    /**
     * Creates a mosaic art spreadsheet from an image file, with
     * mosaic cell size specified.
     *
     * @param file       the source image file
     * @param sheetIndex the index of the mosaic spreadsheet
     * @param maxRows    the maximum number of rows to use, 0 means
     *                   to calculate automatically
     * @throws java.io.IOException when file couldn't be found or
     *         was corrupt
     */
    public void createMosaic(File file, int sheetIndex, int maxRows)
            throws java.io.IOException {
        BufferedImage image;
        
        try {
            image = ImageIO.read(file);
        } catch (java.io.IOException e) {
            throw e;
        }
        createMosaic(image, sheetIndex, file.getName(), maxRows);
    }
    
    /**
     * Creates a mosaic art spreadsheet from an image, with
     * mosaic cell size calculated automatically.
     *
     * @param image      the source image
     * @param sheetIndex the index of the mosaic spreadsheet
     * @param sheetName  the name of the mosaic spreadsheet
     */
    public void createMosaic(BufferedImage image, int sheetIndex,
            String sheetName) {
        createMosaic(image, sheetIndex, sheetName, 0);
    }
    
    /**
     * Creates a mosaic art spreadsheet from an image, with
     * mosaic cell size specified.
     *
     * @param image      the source image
     * @param sheetIndex the index of the mosaic spreadsheet
     * @param sheetName  the name of the mosaic spreadsheet
     * @param maxRows    the maximum number of rows to use, 0 means
     *                   to calculate automatically
     */
    public void createMosaic(BufferedImage image, int sheetIndex,
            String sheetName, int maxRows) {
        if (sheetIndex == 0) {
            createMosaic(image, sheetIndex, sheetName, maxRows, true);
        } else {
            createMosaic(image, sheetIndex, sheetName, maxRows, false);
        }
    }
    
    /**
     * Creates a mosaic art spreadsheet from an image.
     *
     * @param image      the source image
     * @param sheetIndex the index of the mosaic spreadsheet
     * @param sheetName  the name of the mosaic spreadsheet
     * @param maxRows    the maximum number of rows to use, 0 means
     *                   to calculate automatically
     * @param isCleanUp  whether we should remove all the other
     *                   existing spreadsheets
     */
    public void createMosaic(BufferedImage image, int sheetIndex,
            String sheetName, int maxRows, boolean isCleanUp) {
        XSpreadsheet sheet = null;
        XPropertySet xProps = null;
        int cellHeight = -1;
        int cellSize = -1;
        
        sheet = this.setMosaicSheet(sheetIndex, sheetName, isCleanUp);
        cellHeight = this.getSheetCellHeight(sheet);
        cellSize = this.getMosaicCellSize(maxRows, image, cellHeight);
        
        // Set the cell height and width to the same size
        this.setSheetCellWidth(sheet, cellHeight, image, cellSize);
        
        // Draw each cell
        for (int y0 = 0; y0 < image.getHeight(); y0 += cellSize) {
            int y1 = y0 + cellSize - 1;
            if (y1 >= image.getHeight()) {
                y1 = image.getHeight() - 1;
            }
            for (int x0 = 0; x0 < image.getWidth(); x0 += cellSize) {
                int x1 = x0 + cellSize - 1;
                if (x1 >= image.getWidth()) {
                    x1 = image.getWidth() - 1;
                }
                Color color = getMosaicColor(image, x0, y0, x1, y1);
                XCell cell = null;
                try {
                    cell = sheet.getCellByPosition(
                        x0 / cellSize, y0 / cellSize);
                } catch (com.sun.star.lang.IndexOutOfBoundsException e) {
                    throw new java.lang.IndexOutOfBoundsException(
                        e.getMessage());
                }
                xProps = (XPropertySet) UnoRuntime.queryInterface(
                    XPropertySet.class, cell);
                try {
                    xProps.setPropertyValue("CellBackColor", color.getRGB());
                    xProps.setPropertyValue("IsCellBackgroundTransparent", false);
                } catch (com.sun.star.beans.UnknownPropertyException e) {
                    throw new java.lang.RuntimeException(e);
                } catch (com.sun.star.beans.PropertyVetoException e) {
                    throw new java.lang.RuntimeException(e);
                } catch (com.sun.star.lang.IllegalArgumentException e) {
                    throw new java.lang.IllegalArgumentException(e);
                } catch (com.sun.star.lang.WrappedTargetException e) {
                    throw new java.lang.RuntimeException(e);
                }
            }
        }
    }
    
    /**
     * Sets and returns the sheet for our mosaic art.
     *
     * @param sheetIndex the index of the mosaic spreadsheet
     * @param sheetName  the name of the mosaic spreadsheet
     * @param isCleanUp  whether we should remove all the other
     *                   existing spreadsheets
     */
    private XSpreadsheet setMosaicSheet(int sheetIndex,
            String sheetName, boolean isCleanUp) {
        XSpreadsheets sheets = this.document.getSheets();
        Object sheet = null;
        XSpreadsheet xSheet = null;
        
        sheets.insertNewByName(sheetName, (short) sheetIndex);
        try {
            sheet = sheets.getByName(sheetName);
        } catch (com.sun.star.container.NoSuchElementException e) {
            throw new java.lang.RuntimeException(e);
        } catch (com.sun.star.lang.WrappedTargetException e) {
            throw new java.lang.RuntimeException(e);
        }
        
        if (isCleanUp) {
            this.cleanUpOtherSheets(sheetName);
        }
        
        xSheet = (XSpreadsheet) UnoRuntime.queryInterface(
            XSpreadsheet.class, sheet);
        this.xSpreadsheetView.setActiveSheet(xSheet);
        return xSheet;
    }
    
    /**
     * Cleans-up other spreadsheets than the specified one.
     *
     * @param keptSheet  the name of the spreadsheet to keep
     */
    private void cleanUpOtherSheets(String keptSheet) {
        XSpreadsheets sheets = this.document.getSheets();
        
        for (String name : sheets.getElementNames()) {
            if (!name.equals(keptSheet)) {
                try {
                    sheets.removeByName(name);
                } catch (com.sun.star.container.NoSuchElementException e) {
                    throw new java.lang.RuntimeException(e);
                } catch (com.sun.star.lang.WrappedTargetException e) {
                    throw new java.lang.RuntimeException(e);
                }
            }
        }
        return;
    }
    
    /**
     * Returns the mosaic cell size.
     * 
     * @param maxRows    the maximum number of rows to use, 0 means
     *                   to calculate automatically
     * @param image      the source image
     * @param cellHeight the height of the spreadsheet cell
     * @return the calculated mosaic cell size
     */
    private int getMosaicCellSize(int maxRows, BufferedImage image,
            int cellHeight) {
        if (maxRows != 0) {
            return (image.getHeight() - 1) / maxRows + 1;
        }
        
        Dimension windowSize = this.getWindowSize();
        int numColumns = windowSize.width * 25 / cellHeight;
        int numRows = windowSize.height * 25 / cellHeight;
        int mosaicWidth = (image.getWidth() - 1) / numColumns + 1;
        int mosaicHeight = (image.getHeight() - 1) / numRows + 1;
        return mosaicWidth > mosaicHeight? mosaicWidth: mosaicHeight;
    }
    
    /**
     * Returns the size of the window.
     *
     * @return the size of the window
     */
    private Dimension getWindowSize() {
        XFrame frame = this.xController.getFrame();
        XWindow window = frame.getComponentWindow();
        XView xView = (XView) UnoRuntime.queryInterface(
            XView.class, window);
        Size size = xView.getSize();
        return new Dimension(size.Width - 55, size.Height - 72);
    }
    
    /**
     * Returns the height of the spreadsheet cell.  We do not need
     * the width of the spreadsheet cell, since normally width
     * is always longer than height.
     * 
     * @param sheet the spreadsheet
     * @return the height of the spreadsheet cell
     */
    private int getSheetCellHeight(XSpreadsheet sheet) {
        XColumnRowRange xColumnRowRange = (XColumnRowRange)
            UnoRuntime.queryInterface(
            XColumnRowRange.class, sheet);
        XTableRows rows = xColumnRowRange.getRows();
        Object row = null;
        XPropertySet xProps = null;
        int height = -1;
        
        // Obtains the height of the first row.
        try {
            row = rows.getByIndex(0);
        } catch (com.sun.star.lang.IndexOutOfBoundsException e) {
            throw new java.lang.IndexOutOfBoundsException(
                e.getMessage());
        } catch (com.sun.star.lang.WrappedTargetException e) {
            throw new java.lang.RuntimeException(e);
        }
        xProps = (XPropertySet) UnoRuntime.queryInterface(
            XPropertySet.class, row);
        try {
            height = (Integer) xProps.getPropertyValue("Height");
        } catch (com.sun.star.beans.UnknownPropertyException e) {
            throw new java.lang.RuntimeException(e);
        } catch (com.sun.star.lang.WrappedTargetException e) {
            throw new java.lang.RuntimeException(e);
        }
        
        return height;
    }
    
    /**
     * Sets the width of the spreadsheet cell.
     * 
     * @param sheet    the spreadsheet
     * @param width    the width of the spreadsheet cell
     * @param image    the source image
     * @param cellSize the size of the mosaic cell
     */
    private void setSheetCellWidth(XSpreadsheet sheet, int width,
            BufferedImage image, int cellSize) {
        XColumnRowRange xColumnRowRange = (XColumnRowRange)
            UnoRuntime.queryInterface(
            XColumnRowRange.class, sheet);
        XTableColumns columns = xColumnRowRange.getColumns();
        
        for (int x0 = 0; x0 < image.getWidth(); x0 += cellSize) {
            Object column = null;
            XPropertySet xProps = null;
            
            try {
                column = columns.getByIndex(x0 / cellSize);
            } catch (com.sun.star.lang.IndexOutOfBoundsException e) {
                throw new java.lang.IndexOutOfBoundsException(
                    e.getMessage());
            } catch (com.sun.star.lang.WrappedTargetException e) {
                throw new java.lang.RuntimeException(e);
            }
            xProps = (XPropertySet) UnoRuntime.queryInterface(
                XPropertySet.class, column);
            try {
                xProps.setPropertyValue("Width", width);
            } catch (com.sun.star.beans.UnknownPropertyException e) {
                throw new java.lang.RuntimeException(e);
            } catch (com.sun.star.beans.PropertyVetoException e) {
                throw new java.lang.RuntimeException(e);
            } catch (com.sun.star.lang.IllegalArgumentException e) {
                throw new java.lang.IllegalArgumentException(e);
            } catch (com.sun.star.lang.WrappedTargetException e) {
                throw new java.lang.RuntimeException(e);
            }
        }
        return;
    }
    
    /**
     * Obtain the mosaic color of a block in the image.  This calls
     * the default linear distance weakening.
     * 
     * @param image the image
     * @param x0 the x-coordinate of the image's upper-left corner
     * @param y0 the y-coordinate of the image's upper-left corner
     * @param x1 the x-coordinate of the image's lower-right corner
     * @param y1 the y-coordinate of the image's lower-right corner
     * @param the mosaic color of this block
     */
    private Color getMosaicColor(BufferedImage image,
            int x0, int y0, int x1, int y1) {
        return this.getMosaicColorLinear(image,
            x0, y0, x1, y1);
    }
    
    /**
     * Obtain the mosaic color of a block in the image, using
     * trigonometric distance weakening.
     * 
     * @param image the image
     * @param x0 the x-coordinate of the image's upper-left corner
     * @param y0 the y-coordinate of the image's upper-left corner
     * @param x1 the x-coordinate of the image's lower-right corner
     * @param y1 the y-coordinate of the image's lower-right corner
     * @param the mosaic color of this block
     */
    private Color getMosaicColorTrigonometric(BufferedImage image,
            int x0, int y0, int x1, int y1) {
        double count = 0, alpha = 0, red = 0, green = 0, blue = 0;
        double xc = (double) (x0 + x1) / 2;
        double yc = (double) (y0 + y1) / 2;
        double radius = Math.sqrt((xc - x0) * (xc - x0)
            + (yc - y0) * (yc - y0)) + 1;
        
        for (int x = x0; x <= x1; x++) {
            for (int y = y0; y <= y1; y++) {
                Color color = new Color(image.getRGB(x, y));
                double distance = Math.sqrt((x - xc) * (x - xc)
                    + (y - yc) * (y - yc));
                double power = Math.sin(Math.acos((radius - distance) / radius));
                
                alpha += color.getAlpha() * power;
                red += color.getRed() * power;
                green += color.getGreen() * power;
                blue += color.getBlue() * power;
                count += power;
            }
        }
        
        alpha /= count;
        red /= count;
        green /= count;
        blue /= count;
        return new Color((int) red, (int) green, (int) blue,
            (int) alpha);
    }
    
    /**
     * Obtain the mosaic color of a block in the image, using
     * linear distance weakening.
     * 
     * @param image the image
     * @param x0 the x-coordinate of the image's upper-left corner
     * @param y0 the y-coordinate of the image's upper-left corner
     * @param x1 the x-coordinate of the image's lower-right corner
     * @param y1 the y-coordinate of the image's lower-right corner
     * @param the mosaic color of this block
     */
    private Color getMosaicColorLinear(BufferedImage image,
            int x0, int y0, int x1, int y1) {
        double count = 0, alpha = 0, red = 0, green = 0, blue = 0;
        double xc = (double) (x0 + x1) / 2;
        double yc = (double) (y0 + y1) / 2;
        double radius = Math.sqrt((xc - x0) * (xc - x0)
            + (yc - y0) * (yc - y0)) + 1;
        
        for (int x = x0; x <= x1; x++) {
            for (int y = y0; y <= y1; y++) {
                Color color = new Color(image.getRGB(x, y));
                double distance = Math.sqrt((x - xc) * (x - xc)
                    + (y - yc) * (y - yc));
                double power = (radius - distance) / radius;
                
                alpha += color.getAlpha() * power;
                red += color.getRed() * power;
                green += color.getGreen() * power;
                blue += color.getBlue() * power;
                count += power;
            }
        }
        
        alpha /= count;
        red /= count;
        green /= count;
        blue /= count;
        return new Color((int) red, (int) green, (int) blue,
            (int) alpha);
    }
    
    /**
     * Obtain the mosaic color of a block in the image, using simple
     * average.
     * 
     * @param image the image
     * @param x0 the x-coordinate of the image's upper-left corner
     * @param y0 the y-coordinate of the image's upper-left corner
     * @param x1 the x-coordinate of the image's lower-right corner
     * @param y1 the y-coordinate of the image's lower-right corner
     * @param the mosaic color of this block
     */
    private Color getMosaicColorAverage(BufferedImage image,
            int x0, int y0, int x1, int y1) {
        int count = 0, alpha = 0, red = 0, green = 0, blue = 0;
        
        for (int x = x0; x <= x1; x++) {
            for (int y = y0; y <= y1; y++) {
                Color color = new Color(image.getRGB(x, y));
                alpha += color.getAlpha();
                red += color.getRed();
                green += color.getGreen();
                blue += color.getBlue();
                count++;
            }
        }
        
        alpha /= count;
        red /= count;
        green /= count;
        blue /= count;
        return new Color(red, green, blue, alpha);
    }
    
    /**
     * Gets a string for the given key from this resource bundle
     * or one of its parents.  If the key is missing, returns
     * the key itself.
     * 
     * @param key the key for the desired string 
     * @return the string for the given key 
     */
    private String _(String key) {
        try {
            return this.l10n.getString(key);
        } catch (java.util.MissingResourceException e) {
            return key;
        }
    }
}
